from multiprocessing import Process, Event
import logging
import zmq
import sys
from termcolor import colored
import json
# from .helper import set_logger

# 3s for timeout
REQUEST_TIMEOUT = 3000  # 2500
REQUEST_RETRIES = 3
SERVER_ENDPOINT = "tcp://localhost:5555"
logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)


class Client:

    def __init__(self) -> None:
        self.context = zmq.Context()

        logging.info("Connecting to server…")
        self.client = self.context.socket(zmq.REQ)
        self.client.connect(SERVER_ENDPOINT)

    def send_request(self, request):
        self.client.send(request)

        retries_left = REQUEST_RETRIES
        while True:
            
            if (self.client.poll(REQUEST_TIMEOUT) & zmq.POLLIN) != 0:
                reply = self.client.recv()
                # logging.info("Server replied msg (%s)", reply.decode())
                return reply
            
            retries_left -= 1
            logging.warning("No response from server")
            # Socket is confused. Close and remove it.
            self.client.setsockopt(zmq.LINGER, 0)
            self.client.close()
            if retries_left == 0:
                logging.error("Server seems to be offline, abandoning")
                raise Exception("Server seems to be offline, abandoning")
                # sys.exit()

            logging.info("Reconnecting to server…")
            # Create new connection
            self.client = self.context.socket(zmq.REQ)
            self.client.connect(SERVER_ENDPOINT)
            logging.info("Resending (%s)", request)
            self.client.send(request)

    def close(self):
        """
            Gently close all connections of the client. 
        """
        self.client.close()
        self.context.term()

class BCManager():
    def __init__(self, available_bc):
        self.available_bc = available_bc
        self.bc = None

    def __enter__(self):
        self.bc = self.available_bc.pop()
        return self.bc

    def __exit__(self, *args):
        self.available_bc.append(self.bc)

class ConcurrentClient:

    def __init__(self) -> None:
        max_concurrency = 2
        self.max_concurrency = max_concurrency
        self.available_bc = [Client() for _ in range(max_concurrency)]
        
    def send_request(self, request):
        with BCManager(self.available_bc) as bc:
            return bc.send_request(request)
  

    def close(self):
        for bc in self.available_bc:
            bc.close()


class HTTPProxy(Process):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.is_ready = Event()

    def create_fastapi_app(self):

        from fastapi import FastAPI, Request
        from flask_json import JsonError

        app = FastAPI()

        # logger = set_logger(colored('PROXY', 'red'), self.args.verbose)

        # @app.get('/status/server')
        # def get_server_status():
        #     return bc.server_status

        # @app.get('/status/client')
        # def get_client_status():
        #     return bc.status

        @app.post('/predict')
        def predict(data: dict, request: Request):
            # data = request.form if request.form else request.json
            try:
                print(request, data)
                json_str = json.dumps(data)
                # logger.info('new request from %s' % request.client.host)
                result = self.concurrent.send_request(json_str.encode())
                return result
            except Exception as e:
                print(e)
                # logger.error('error when handling HTTP request', exc_info=True)
                raise JsonError(description=str(e), type=str(type(e).__name__))

        return app

    def run(self):
        import uvicorn
        app = self.create_fastapi_app()
        self.concurrent = ConcurrentClient()
        self.is_ready.set()
        uvicorn.run(app, host="0.0.0.0", port=self.args.http_port)



import zmq

LRU_READY = "\x01"

context = zmq.Context(1)

frontend = context.socket(zmq.ROUTER) # ROUTER
backend = context.socket(zmq.DEALER) # DEALER
frontend.bind("tcp://*:5555") # For clients
backend.bind("tcp://*:5556")  # For workers

zmq.device(zmq.QUEUE, frontend, backend)




#
#  Lazy Pirate client
#  Use zmq_poll to do a safe request-reply
#  To run, start lpserver and then randomly kill/restart it
#
#   Author: Daniel Lundin <dln(at)eintr(dot)org>
#
import itertools
import logging
import sys
import zmq
import time, json
logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)

# 3s for timeout
REQUEST_TIMEOUT = 10000 #2500 
REQUEST_RETRIES = 3
SERVER_ENDPOINT = "tcp://localhost:5555"

context = zmq.Context()

logging.info("Connecting to server…")
client = context.socket(zmq.REQ)
client.connect(SERVER_ENDPOINT)

def send_request(client, request):
    client.send(request)

    retries_left = REQUEST_RETRIES
    while True:
        
        if (client.poll(REQUEST_TIMEOUT) & zmq.POLLIN) != 0:
            reply = client.recv()
            logging.info("Server replied msg (%s)", reply.decode())
            return reply
        
        retries_left -= 1
        logging.warning("No response from server")
        # Socket is confused. Close and remove it.
        client.setsockopt(zmq.LINGER, 0)
        client.close()
        if retries_left == 0:
            logging.error("Server seems to be offline, abandoning")
            sys.exit()

        logging.info("Reconnecting to server…")
        # Create new connection
        client = context.socket(zmq.REQ)
        client.connect(SERVER_ENDPOINT)
        logging.info("Resending (%s)", request)
        client.send(request)

for sequence in itertools.count():
    sequence = {'str': '今天吃饭了吗'}
    sequence = json.dumps(sequence)
    request = sequence.encode()
    logging.info("Sending (%s)", request)
    send_request(client, request)


from typing import List

import aicmder as cmder
from aicmder.commands import register
from aicmder.commands.utils import _command_prefix as cmd
import argparse
import os
from shutil import copyfile
from pprint import pprint
import json
# from multiprocessing import set_start_method
# try:
#      set_start_method('spawn')
# except RuntimeError:
#     pass
import subprocess
def is_json(config):
    if config is None:
        return False
    try:
        json_object = json.loads(config)
    except ValueError as e:
        return False
    return True


@register(name='{}.serve'.format(cmd), description='start all module')
class ServeCommand:

    def __init__(self) -> None:
        self.parser = argparse.ArgumentParser(
            description=self.__class__.__doc__, prog='{} serve'.format(cmd), usage='%(prog)s', add_help=True)
        self.parser.add_argument('--workers', '-w', type=int, default=1, help='number of server instances')
        self.parser.add_argument('--config', '-c', required=False)
        self.parser.add_argument('--file', '-f', required=False)
        self.parser.add_argument('--http_port', '-p', type=int, default=None,
                        help='server port for receiving HTTP requests')
        self.parser.add_argument('-device_map', type=int, nargs='+', default=[],
                                 help='specify the list of GPU device ids that will be used (id starts from 0). \
                        If num_worker > len(device_map), then device will be reused; \
                        if num_worker < len(device_map), then device_map[:num_worker] will be used')
        self.parser.add_argument('-cpu', action='store_true', default=False,
                                 help='running on CPU (default on GPU)')
        base_dir = os.path.dirname(__file__)
        self.work_file = os.path.join(base_dir, '../service','worker.py')
        assert os.path.exists(self.work_file)
        
    def _get_device_map(self):
        # self.logger.info('get devices')
        run_on_gpu = False
        device_map = [-1] * self.num_worker
        if not self.args.cpu:
            try:
                import GPUtil
                num_all_gpu = len(GPUtil.getGPUs())
                avail_gpu = GPUtil.getAvailable(order='memory', limit=min(num_all_gpu, self.num_worker),
                                                maxMemory=0.9, maxLoad=0.9)
                num_avail_gpu = len(avail_gpu)

                if num_avail_gpu >= self.num_worker:
                    run_on_gpu = True
                elif 0 < num_avail_gpu < self.num_worker:
                    # self.logger.warning('only %d out of %d GPU(s) is available/free, but "-num_worker=%d"' %
                    #                     (num_avail_gpu, num_all_gpu, self.num_worker))
                    # if not self.args.device_map:
                    #     self.logger.warning('multiple workers will be allocated to one GPU, '
                    #                         'may not scale well and may raise out-of-memory')
                    # else:
                    #     self.logger.warning('workers will be allocated based on "-device_map=%s", '
                    #                         'may not scale well and may raise out-of-memory' % self.args.device_map)
                    run_on_gpu = True
                else:
                    print('no GPU available, fall back to CPU')
                    # self.logger.warning('no GPU available, fall back to CPU')

                if run_on_gpu:
                    device_map = ((self.args.device_map or avail_gpu) * self.num_worker)[: self.num_worker]
            except FileNotFoundError:
                print('nvidia-smi is missing, often means no gpu on this machine. '
                                    'fall back to cpu!')
        print('device map: \n\t\t%s' % '\n\t\t'.join(
            'worker %2d -> %s' % (w_id, ('gpu %2d' % g_id) if g_id >= 0 else 'cpu') for w_id, g_id in
            enumerate(device_map)))
        return device_map

    def execute(self, argv: List) -> bool:
        print(argv)
        self.args = self.parser.parse_args(argv)
        self.num_worker = self.args.workers
        device_map = self._get_device_map()
        print(device_map)
        
        assert is_json(self.args.config) == True or os.path.exists(self.args.file) == True
        if self.args.file is not None and os.path.exists(self.args.file):
            with open(self.args.file, 'r') as f:
                content = f.read()
                config = json.loads(content)
        else:
            config = json.loads(self.args.config)
    
        workers = []
        for idx, device_id in enumerate(device_map):
            subprocess.Popen(['python', self.work_file, json.dumps(config), str(device_id)])
            
            # worker = cmder.Worker(config, device_id=device_id)
            # workers.append(worker)
            # worker.start()
        if self.args.http_port:
            # self.logger.info('start http proxy')
            proc_proxy = cmder.HTTPProxy(self.args)
            workers.append(proc_proxy)
            proc_proxy.start()    
        
        for worker in workers:
            worker.is_ready.wait()
        queue = cmder.ServerQueue()
        queue.start()
        queue.join()
        



#
##  Paranoid Pirate worker
#
#   Author: Daniel Lundin <dln(at)eintr(dot)org>
#

from multiprocessing.context import Process
from random import randint
import time
from aicmder.service.PPpolicy import *
import zmq



def worker_socket(context, poller):
    """Helper function that returns a new configured socket
       connected to the Paranoid Pirate queue"""
    worker = context.socket(zmq.DEALER) # DEALER
    identity = b"%04X-%04X" % (randint(0, 0x10000), randint(0, 0x10000))
    worker.setsockopt(zmq.IDENTITY, identity)
    poller.register(worker, zmq.POLLIN)
    worker.connect("tcp://localhost:5556")
    worker.send(PPP_READY)
    return worker

class PPworker(Process):

    def run(self):    
        context = zmq.Context(1)
        poller = zmq.Poller()

        liveness = HEARTBEAT_LIVENESS
        interval = INTERVAL_INIT

        heartbeat_at = time.time() + HEARTBEAT_INTERVAL

        worker = worker_socket(context, poller)
        cycles = 0
        while True:
            socks = dict(poller.poll(HEARTBEAT_INTERVAL * 1000))

            # Handle worker activity on backend
            if socks.get(worker) == zmq.POLLIN:
                #  Get message
                #  - 3-part envelope + content -> request
                #  - 1-part HEARTBEAT -> heartbeat
                frames = worker.recv_multipart()
                if not frames:
                    break # Interrupted

                if len(frames) == 3:
                    # Simulate various problems, after a few cycles
                    cycles += 1
                    if cycles > 3 and randint(0, 5) == 0:
                        print("I: Simulating a crash")
                        break
                    if cycles > 3 and randint(0, 5) == 0:
                        print("I: Simulating CPU overload")
                        time.sleep(3)
                    print("I: Normal reply")
                    worker.send_multipart(frames)
                    liveness = HEARTBEAT_LIVENESS
                    time.sleep(1)  # Do some heavy work
                elif len(frames) == 1 and frames[0] == PPP_HEARTBEAT:
                    print("I: Queue heartbeat")
                    liveness = HEARTBEAT_LIVENESS
                else:
                    print("E: Invalid message: %s" % frames)
                interval = INTERVAL_INIT
            else:
                liveness -= 1
                if liveness == 0:
                    print("W: Heartbeat failure, can't reach queue")
                    print("W: Reconnecting in %0.2fs..." % interval)
                    time.sleep(interval)

                    if interval < INTERVAL_MAX:
                        interval *= 2
                    poller.unregister(worker)
                    worker.setsockopt(zmq.LINGER, 0)
                    worker.close()
                    worker = worker_socket(context, poller)
                    liveness = HEARTBEAT_LIVENESS
            if time.time() > heartbeat_at:
                heartbeat_at = time.time() + HEARTBEAT_INTERVAL
                print("I: Worker heartbeat")
                worker.send(PPP_HEARTBEAT)


from typing import Callable
import inspect
from aicmder.utils.utils import load_py_module, Version, load_py_dir
import os
_module_serving_func = {}
_module_runnable_func = {}


class InvalidModule(Exception):
    def __init__(self, directory: str):
        self.directory = directory

    def __str__(self):
        return '{} is not a valid Module'.format(self.directory)


def runnable(func: Callable) -> Callable:
    '''Mark a Module method as runnable, when the command `hub run` is used, the method will be called.'''
    mod = func.__module__ + '.' + inspect.stack()[1][3]
    _module_runnable_func[mod] = func.__name__

    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return _wrapper


def serving(func: Callable) -> Callable:
    '''Mark a Module method as serving method, when the command `hub serving` is used, the method will be called.'''
    mod = func.__module__ + '.' + inspect.stack()[1][3]
    _module_serving_func[mod] = func.__name__

    def _wrapper(*args, **kwargs):
        func_signature = inspect.signature(func)
        if len(func_signature.parameters) <= 1:
            print(args, kwargs)
            return func(*args)
        return func(*args, **kwargs)

    return _wrapper


class Module(object):
    '''
    In PaddleHub, Module represents an executable module, which usually a pre-trained model that can be used for end-to-end
    prediction, such as a face detection model or a lexical analysis model, or a pre-trained model that requires finetuning,
    such as BERT/ERNIE. When loading a Module with a specified name, if the Module does not exist locally, PaddleHub will
    automatically request the server or the specified Git source to download the resource.

    Args:
        name(str): Module name.
        directory(str|optional): Directory of the module to be loaded, only takes effect when the `name` is not specified.
        version(str|optional): The version limit of the module, only takes effect when the `name` is specified. When the local
                               Module does not meet the specified version conditions, PaddleHub will re-request the server to
                               download the appropriate Module. Default to None, This means that the local Module will be used.
                               If the Module does not exist, PaddleHub will download the latest version available from the
                               server according to the usage environment.
        source(str|optional): Url of a git repository. If this parameter is specified, PaddleHub will no longer download the
                              specified Module from the default server, but will look for it in the specified repository.
                              Default to None.
        update(bool|optional): Whether to update the locally cached git repository, only takes effect when the `source`
                               is specified. Default to False.
        branch(str|optional): The branch of the specified git repository. Default to None.
    '''

    def __init__(self, **kwargs) -> None:
        pass

    def __new__(cls,
                *,
                name: str = None,
                directory: str = None,
                version: str = None,
                source: str = None,
                update: bool = False,
                branch: str = None,
                **kwargs):
        if cls.__name__ == 'Module':
            # from paddlehub.server.server import CacheUpdater
            # This branch come from hub.Module(name='xxx') or hub.Module(directory='xxx')
            if name:
                module = cls.init_with_name(
                    name=name, version=version, source=source, update=update, branch=branch, **kwargs)
                # CacheUpdater("update_cache", module=name, version=version).start()
            elif directory:
                module = cls.init_with_directory(directory=directory, **kwargs)
                # CacheUpdater("update_cache", module=directory, version="0.0.0").start()
        else:
            module = object.__new__(cls)

        return module

    def _get_func_name(self, current_cls, module_func_dict: dict):
        mod = current_cls.__module__ + '.' + current_cls.__name__
        if mod in module_func_dict:
            _func_name = module_func_dict[mod]
            return _func_name
        elif current_cls.__bases__:
            for base_class in current_cls.__bases__:
                base_run_func = self._get_func_name(base_class, module_func_dict)
                if base_run_func:
                    return base_run_func
        else:
            return None

    @property
    def serving_func_name(self):
        return self._get_func_name(self.__class__, _module_serving_func)

    @classmethod
    def load(cls, directory: str):
        '''Load the Module object defined in the specified directory.'''
        if directory.endswith(os.sep):
            directory = directory[:-1]

        # try:
        #     # check class exists or not
        #     py_module = eval(directory)()
        # except:
        basename = os.path.basename(directory)
        dirname = os.path.dirname(directory)
        load_py_dir(os.path.join(os.getcwd(), dirname))
        py_module = load_py_module(dirname, basename)

        for _item, _cls in inspect.getmembers(py_module, inspect.isclass):
            _item = py_module.__dict__[_item]
            if hasattr(_item, '_hook_by_cmder') and issubclass(_item, Module):
                user_module_cls = _item
                break
        else:
            raise InvalidModule(directory)

        user_module_cls.directory = directory

        # source_info_file = os.path.join(directory, '_source_info.yaml')
        # if os.path.exists(source_info_file):
        #     info = parser.yaml_parser.parse(source_info_file)
        #     user_module_cls.source = info.get('source', '')
        #     user_module_cls.branch = info.get('branch', '')
        # else:
        #     user_module_cls.source = ''
        #     user_module_cls.branch = ''

        # # In the case of multiple cards, the following code can set each process to use the correct place.
        # if issubclass(user_module_cls, paddle.nn.Layer):
        #     place = paddle.get_device().split(':')[0]
        #     paddle.set_device(place)

        return user_module_cls

    @classmethod
    def init_with_name(cls,
                       name: str,
                       version: str = None,
                       source: str = None,
                       update: bool = False,
                       branch: str = None,
                       **kwargs):
        '''Initialize Module according to the specified name.'''
        # from paddlehub.module.manager import LocalModuleManager
        # manager = LocalModuleManager()
        # user_module_cls = manager.search(name, source=source, branch=branch)
        # if not user_module_cls or not user_module_cls.version.match(version):
        #     user_module_cls = manager.install(
        #         name=name, version=version, source=source, update=update, branch=branch)

        # directory = manager._get_normalized_path(user_module_cls.name)

        # # The HubModule in the old version will use the _initialize method to initialize,
        # # this function will be obsolete in a future version
        # if hasattr(user_module_cls, '_initialize'):
        #     log.logger.warning(
        #         'The _initialize method in HubModule will soon be deprecated, you can use the __init__() to handle the initialization of the object'
        #     )
        #     user_module = user_module_cls(directory=directory)
        #     user_module._initialize(**kwargs)
        #     return user_module

        # if issubclass(user_module_cls, ModuleV1):
        #     return user_module_cls(directory=directory, **kwargs)

        # user_module_cls.directory = directory
        user_module_cls = cls.load(name)
        return user_module_cls(**kwargs)


def moduleinfo(name: str = '',
               version: str = '1.0',
               author: str = None,
               author_email: str = None,
               summary: str = None,
               type: str = None,
               meta=None) -> Callable:
    '''
    Mark Module information for a python class, and the class will automatically be extended to inherit HubModule. In other words, python classes
    marked with moduleinfo can be loaded through hub.Module.
    '''

    def _wrapper(cls):
        wrap_cls = cls
        _meta = Module if not meta else meta
        if not issubclass(cls, _meta):
            _bases = []
            for _b in cls.__bases__:
                if issubclass(_meta, _b):
                    continue
                _bases.append(_b)
            _bases.append(_meta)
            _bases = tuple(_bases)
            wrap_cls = builtins.type(cls.__name__, _bases, dict(cls.__dict__))

        wrap_cls.name = name
        wrap_cls.version = Version(version)
        wrap_cls.author = author
        wrap_cls.author_email = author_email
        wrap_cls.summary = summary
        wrap_cls.type = type
        wrap_cls._hook_by_cmder = True
        return wrap_cls

    return _wrapper



#
##  Paranoid Pirate queue
#
#   Author: Daniel Lundin <dln(at)eintr(dot)org>
#

from collections import OrderedDict
import time
from PPpolicy import *
import zmq




class Worker(object):
    def __init__(self, address):
        self.address = address
        self.expiry = time.time() + HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS

class WorkerQueue(object):
    def __init__(self):
        self.queue = OrderedDict()

    def ready(self, worker):
        self.queue.pop(worker.address, None)
        self.queue[worker.address] = worker

    def purge(self):
        """Look for & kill expired workers."""
        t = time.time()
        expired = []
        for address, worker in self.queue.items():
            if t > worker.expiry:  # Worker expired
                expired.append(address)
        for address in expired:
            print("W: Idle worker expired: %s" % address)
            self.queue.pop(address, None)

    def next(self):
        address, worker = self.queue.popitem(False)
        return address

context = zmq.Context(1)

frontend = context.socket(zmq.ROUTER) # ROUTER
backend = context.socket(zmq.ROUTER)  # ROUTER
frontend.bind("tcp://*:5555") # For clients
backend.bind("tcp://*:5556")  # For workers

poll_workers = zmq.Poller()
poll_workers.register(backend, zmq.POLLIN)

poll_both = zmq.Poller()
poll_both.register(frontend, zmq.POLLIN)
poll_both.register(backend, zmq.POLLIN)

workers = WorkerQueue()

heartbeat_at = time.time() + HEARTBEAT_INTERVAL

while True:
    if len(workers.queue) > 0:
        poller = poll_both
    else:
        poller = poll_workers
    socks = dict(poller.poll(HEARTBEAT_INTERVAL * 1000))

    # Handle worker activity on backend
    if socks.get(backend) == zmq.POLLIN:
        # Use worker address for LRU routing
        frames = backend.recv_multipart()
        if not frames:
            break

        address = frames[0]
        workers.ready(Worker(address))

        # Validate control message, or return reply to client
        msg = frames[1:]
        if len(msg) == 1:
            if msg[0] not in (PPP_READY, PPP_HEARTBEAT):
                print("E: Invalid message from worker: %s" % msg)
        else:
            frontend.send_multipart(msg)

        # Send heartbeats to idle workers if it's time
        if time.time() >= heartbeat_at:
            for worker in workers.queue:
                msg = [worker, PPP_HEARTBEAT]
                backend.send_multipart(msg)
            heartbeat_at = time.time() + HEARTBEAT_INTERVAL
    if socks.get(frontend) == zmq.POLLIN:
        frames = frontend.recv_multipart()
        if not frames:
            break

        frames.insert(0, workers.next())
        backend.send_multipart(frames)


    workers.purge()


from aicmder.utils.utils import load_py_dir
from typing import List

import aicmder as cmder
from aicmder.commands import register
from aicmder.commands.utils import _command_prefix as cmd
from aicmder.commands.utils import help_str
import argparse
from importlib import import_module
import os, sys
from pprint import pprint
import torch
import importlib.util
@register(name='{}.onnx'.format(cmd), description='convert to onnx.')
class ONNXCommand:
    
    _helper = 'export_helper'
    
    def __init__(self) -> None:
        self.description = 'Start to convert to onnx model'
        self.parser = argparse.ArgumentParser(description=self.__class__.__doc__, prog='{} onnx'.format(cmd), usage='%(prog)s', add_help=True)
        self.parser.add_argument('--save_dir', '-s', required=False)
        self.parser.add_argument('--model', '-m', required=True)
        self.parser.add_argument('--ckpt', '-c', required=True)
        self.parser.add_argument('--gpu', action="store_true", default=False)
        self.parser.add_argument('--verbose', action="store_true", default=False)
        self.parser.add_argument('--export', '-e', default='export.onnx')
        self.cur_path = os.path.abspath(os.path.dirname(__file__))
        
        
    def execute(self, argv: List) -> bool:
        pprint('begin to convert to onnx')
        args = self.parser.parse_args(argv)
        save_dir = self.cur_path if args.save_dir is None else args.save_dir
        assert os.path.exists(args.model) and os.path.exists(args.ckpt)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        
        model_name = os.path.basename(args.model)
        model_dir = os.path.dirname(args.model)
        model_basename = os.path.basename(model_dir)
        # sys.path.append(model_dir)
        
        ## import module
        module = load_py_dir(model_dir)

        MODLE = import_module('{}.{}'.format(model_basename, model_name.replace('.py', '')))
        
        device = 'cuda:0' if args.gpu and torch.cuda.is_available() else 'cpu'
        # print(model_dir, module, spec.name, MODLE, args.gpu, device)
        
        ## build model
        export_helper = import_module('{}.{}'.format(model_basename, self._helper))
        model = MODLE.make_model(export_helper.args).to(device)
        state_dicts = torch.load(args.ckpt, map_location=device)
        model.load_state_dict(state_dicts)
        print('load model success.')
        
        ## export model
        dummy_input = export_helper.dummy_input.to(device)
        torch.onnx.export(model, dummy_input,os.path.join(save_dir, args.export),verbose=args.verbose)
        # print(model)
        return True

    @staticmethod
    def help():
        str = "onnx example\n"
        str += help_str("aicmder onnx -s output_dir -m './model/edsr.py' -c ckpt")
        str += help_str("-s onnx model output dir")
        str += help_str("-m py file of model")
        return str


from typing import List


from aicmder.commands import register, _commands
from aicmder.commands.utils import _command_prefix as cmd


@register(name='{}.help'.format(cmd), description='Show help for commands.')
class HelpCommand:
    def execute(self, argv: List) -> bool:
        msg = 'Usage:\n'
        msg += '    {} <command> <options>\n\n'.format(cmd)
        msg += 'Commands:\n'
        for command, detail in _commands[cmd].items():
            if command.startswith('_'):
                continue

            if not '_description' in detail:
                continue
            msg += '    {:<15}        {}\n'.format(command, detail['_description'])
            if hasattr(detail['_entry'], 'help'):
                msg += '    {:<15}        {}\n'.format('', detail['_entry'].help())
        print(msg)
        return True


from model import common

import torch.nn as nn

url = {
    'r16f64x2': 'https://cv.snu.ac.kr/research/EDSR/models/edsr_baseline_x2-1bc95232.pt',
    'r16f64x3': 'https://cv.snu.ac.kr/research/EDSR/models/edsr_baseline_x3-abf2a44e.pt',
    'r16f64x4': 'https://cv.snu.ac.kr/research/EDSR/models/edsr_baseline_x4-6b446fab.pt',
    'r32f256x2': 'https://cv.snu.ac.kr/research/EDSR/models/edsr_x2-0edfb8a3.pt',
    'r32f256x3': 'https://cv.snu.ac.kr/research/EDSR/models/edsr_x3-ea3ef2c6.pt',
    'r32f256x4': 'https://cv.snu.ac.kr/research/EDSR/models/edsr_x4-4f62e9ef.pt'
}

def make_model(args, parent=False):
    return EDSR(args)

class EDSR(nn.Module):
    def __init__(self, args, conv=common.default_conv):
        super(EDSR, self).__init__()

        n_resblocks = args.n_resblocks
        n_feats = args.n_feats
        kernel_size = 3 
        scale = args.scale[0]
        act = nn.ReLU(True)
        url_name = 'r{}f{}x{}'.format(n_resblocks, n_feats, scale)
        if url_name in url:
            self.url = url[url_name]
        else:
            self.url = None
        self.sub_mean = common.MeanShift(args.rgb_range)
        self.add_mean = common.MeanShift(args.rgb_range, sign=1)

        # define head module
        m_head = [conv(args.n_colors, n_feats, kernel_size)]

        # define body module
        m_body = [
            common.ResBlock(
                conv, n_feats, kernel_size, act=act, res_scale=args.res_scale
            ) for _ in range(n_resblocks)
        ]
        m_body.append(conv(n_feats, n_feats, kernel_size))

        # define tail module
        m_tail = [
            common.Upsampler(conv, scale, n_feats, act=False),
            conv(n_feats, args.n_colors, kernel_size)
        ]

        self.head = nn.Sequential(*m_head)
        self.body = nn.Sequential(*m_body)
        self.tail = nn.Sequential(*m_tail)

    def forward(self, x):
        x = self.sub_mean(x)
        x = self.head(x)

        res = self.body(x)
        res += x

        x = self.tail(res)
        x = self.add_mean(x)

        return x 

    def load_state_dict(self, state_dict, strict=True):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            if name in own_state:
                if isinstance(param, nn.Parameter):
                    param = param.data
                try:
                    own_state[name].copy_(param)
                except Exception:
                    if name.find('tail') == -1:
                        raise RuntimeError('While copying the parameter named {}, '
                                           'whose dimensions in the model are {} and '
                                           'whose dimensions in the checkpoint are {}.'
                                           .format(name, own_state[name].size(), param.size()))
            elif strict:
                if name.find('tail') == -1:
                    raise KeyError('unexpected key "{}" in state_dict'
                                   .format(name))

